from .chat import AsyncChatModule, SyncChatModule
from .collections import AsyncCollectionsModule, SyncCollectionsModule
from .embeddings import AsyncEmbeddingsModule, SyncEmbeddingsModule
from .me import AsyncMeModule, SyncMeModule
from .models import AsyncModelsModule, SyncModelsModule
from .workflows import AsyncWorkflowsModule, SyncWorkflowsModule
from .tools import ToolsModule
from .documents import AsyncDocumentsModule, SyncDocumentsModule
from .s3 import SyncS3Module

__all__ = [
    "AsyncChatModule",
    "SyncChatModule",
    "AsyncCollectionsModule",
    "SyncCollectionsModule",
    "AsyncEmbeddingsModule",
    "SyncEmbeddingsModule",
    "AsyncMeModule",
    "SyncMeModule",
    "AsyncModelsModule",
    "SyncModelsModule",
    "AsyncWorkflowsModule",
    "SyncWorkflowsModule",
    "ToolsModule",
    "AsyncDocumentsModule",
    "SyncDocumentsModule",
    "SyncS3Module",
]



from datetime import datetime
from typing import Optional, List, Dict, Any
from pydantic import BaseModel, ConfigDict


class Workflow(BaseModel):
    name: str
    id: str
    group: str
    createdAt: datetime
    updatedAt: datetime
    description: Optional[str] = None
    isActive: Optional[bool] = True


class Input(BaseModel):
    input_id: str
    input_type: str
    input_name: str
    value: Dict[str, Any]
    linkable: bool


class Output(BaseModel):
    output_id: str
    output_type: str
    output_name: str
    linkable: bool


class Position(BaseModel):
    x: float
    y: float
    _id: str


class Node(BaseModel):
    inputs: List[Input]
    outputs: List[Output]
    node_id: str
    name: str
    source_nodes_id: Optional[List[str]]
    type: str
    target_nodes: List[Dict[str, Any]]
    position: Position
    _id: str


class WorkflowDetails(BaseModel):
    id: str
    group: str
    name: str
    createdAt: datetime
    updatedAt: datetime
    starting_node_id: str
    nodes: List[Node]


class WorkflowRunChunk(BaseModel):
    response: Optional[str] = None
    model_config = ConfigDict(extra="allow")


class WorkflowRunResult(BaseModel):
    response: Optional[str] = None
    model_config = ConfigDict(extra="allow")



from pydantic import BaseModel
from typing import Optional, List, Tuple, Union
from io import BytesIO


class DocumentParserResponse(BaseModel):
    """
    Response model for document parsing.
    """

    message: Optional[str] = None
    parsedFiles: Optional[List[str]] = None
    bucketName: Optional[str] = None


class FileObject(BaseModel):
    """
    Represents a file object with its name and content.
    """

    name: str
    buffer: bytes


class DocumentParserDocOutput(BaseModel):
    type: str
    data: Union[str, dict, list]


class ParsedDocument(BaseModel):
    """
    Represents a parsed document with its content and metadata.
    """

    content: DocumentParserDocOutput
    metadata: Optional[dict] = None


DocumentParserFileInput = Union[
    str,
    bytes,
    BytesIO,
    Tuple[bytes, str],
    "FileObject",
    List[Union[str, bytes, BytesIO, Tuple[bytes, str], "FileObject"]],
]



from typing import Type, TypeVar, Any, Dict, Optional, Callable
import httpx
from pydantic import ValidationError, BaseModel

T = TypeVar("T", bound=BaseModel)


async def areq(
    client: httpx.AsyncClient,
    method: str,
    url: str,
    response_model: Type[T],
    *,
    expected_status: int = 200,
    handle_404_none: bool = False,
    **kwargs,
) -> Optional[T]:
    """
    Send an HTTP request asynchronously, parse the response.
    Return None if 404 and handle_404_none=True, else raise.
    """
    try:
        resp = await client.request(method, url, **kwargs)
        if handle_404_none and resp.status_code == 404:
            return None

        if 400 <= resp.status_code < 500:
            resp.read()
            if resp.text == "Please specify a group":
                raise httpx.HTTPStatusError(
                    "Bad Request: Please specify a group. Use `AisbergClient.me.groups()` to know which groups you are in.",
                    request=resp.request,
                    response=resp,
                )
            raise httpx.HTTPStatusError(
                f"Bad Request: {resp.text}", request=resp.request, response=resp
            )

        resp.raise_for_status()
        return response_model.model_validate(resp.json())
    except httpx.HTTPStatusError as e:
        if handle_404_none and e.response.status_code == 404:
            return None
        raise
    except ValidationError as ve:
        raise RuntimeError(f"Invalid response: {ve}") from ve


async def areq_stream(
    client: httpx.AsyncClient,
    method: str,
    url: str,
    parse_line: Callable[[str], Any],
    *,
    handle_status: Optional[Dict[int, Any]] = None,
    **kwargs,
) -> Any:
    """
    Wrapper to handle HTTP streams.
    - parse_line: Function to parse each line of the stream.
    """
    async with client.stream(method, url, **kwargs) as resp:
        if handle_status and resp.status_code in handle_status:
            yield handle_status[resp.status_code]
            return

        if 400 <= resp.status_code < 500:
            resp.read()
            if resp.text == "Please specify a group":
                raise httpx.HTTPStatusError(
                    "Bad Request: Please specify a group. Use `await AisbergAsyncClient.me.groups()` to know which groups you are in.",
                    request=resp.request,
                    response=resp,
                )
            raise httpx.HTTPStatusError(
                f"Bad Request: {resp.text}", request=resp.request, response=resp
            )

        resp.raise_for_status()
        async for raw_line in resp.aiter_lines():
            if not raw_line:
                continue
            line = raw_line.decode() if isinstance(raw_line, bytes) else raw_line
            for parsed in parse_line(line):
                yield parsed



from typing import Iterator, Union, Dict
import json


def parse_chat_line(
    line: str, *, full_chunk: bool = True
) -> Iterator[Union[str, dict]]:
    """
    Parse une ligne de stream JSON (commençant par `data:`) et yield un ou plusieurs éléments utilisables.

    Args:
        line (str): Ligne du flux à traiter.
        full_chunk (bool): Contrôle le format de sortie (chunk brut ou contenu transformé).

    Yields:
        Union[str, dict]: Le chunk complet ou un morceau de texte/fonction/tool_call.
    """
    if not line.startswith("data:"):
        return

    data = line[len("data:") :].strip()
    if data == "[DONE]":
        return

    try:
        chunk: dict = json.loads(data)

        if chunk["object"] != "chat.completion.chunk":
            raise ValueError(f"Unexpected object type: {chunk['object']}")

        if not chunk.get("choices", []):
            return

        if chunk["choices"][0]["finish_reason"] == "stop":
            return

        if full_chunk:
            yield chunk
            return

        delta = chunk.get("choices", [{}])[0].get("delta", {})

        if "content" in delta and delta["content"]:
            yield delta["content"]

        elif "function_call" in delta and delta["function_call"]:
            yield f"[FUNCTION_CALL]{delta['function_call']}"

        elif "tool_calls" in delta and delta["tool_calls"]:
            yield f"[TOOL_CALLS]{delta['tool_calls']}"

    except json.JSONDecodeError:
        return


class WorkflowLineParser:
    def __init__(self, full_chunk: bool = True):
        self.full_chunk = full_chunk
        self._buckets: Dict[str, Dict] = {}

    def __call__(self, line: str) -> Iterator[Union[str, dict]]:
        if not line.startswith("data:"):
            return
        data = line[len("data:") :].strip()
        if data == "[DONE]":
            return

        try:
            payload = json.loads(data)

            # Gestion des events chunk + slices
            if payload.get("type") == "chunk":
                self._buckets[payload["id"]] = {
                    "totalSlices": payload["totalSlices"],
                    "slices": [None] * payload["totalSlices"],
                }

            elif payload.get("type") == "chunk_slice":
                bucket = self._buckets.get(payload["id"])
                if bucket:
                    bucket["slices"][payload["index"]] = payload["slice"]

            elif payload.get("type") == "chunk_end":
                bucket = self._buckets.pop(payload["id"], None)
                if bucket:
                    full_json = "".join(bucket["slices"])
                    try:
                        yield from self._yield_chunk(full_json)
                    except json.JSONDecodeError:
                        return

            else:
                # Message "normal" non splitté
                yield from self._yield_chunk(data)

        except json.JSONDecodeError:
            return

    def _yield_chunk(self, raw: str) -> Iterator[Union[str, dict]]:
        chunk = json.loads(raw)

        if "slice" in chunk:
            try:
                if isinstance(chunk["slice"], str):
                    # Si la slice est déjà un JSON string, on la parse
                    parsed_slice = json.loads(chunk["slice"])
                    yield parsed_slice
                elif isinstance(chunk["slice"], dict):
                    # Si la slice est déjà un dict, on la yield directement
                    yield chunk["slice"]
            except json.JSONDecodeError:
                return
        else:
            return



from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, Union, Optional

if TYPE_CHECKING:
    from ..client import AisbergClient
    from ..async_client import AisbergAsyncClient
    from httpx import Client as HttpClient
    from httpx import AsyncClient as AsyncHttpClient


class BaseModule(ABC):
    """Abstract base class for modules in the Aisberg framework."""

    def __init__(
        self,
        parent: Union["AisbergClient", "AisbergAsyncClient"],
        http_client: Optional[Union["HttpClient", "AsyncHttpClient"]] = None,
    ):
        """
        Initialize the BaseModule.

        Args:
            parent (Any): Parent client or module.
            http_client (Any): HTTP client for making requests.
        """
        self._parent = parent
        self._client = http_client


class SyncModule(BaseModule):
    """Abstract base class for synchronous modules in the Aisberg framework."""

    def __init__(self, parent: "AisbergClient", http_client: "HttpClient"):
        """
        Initialize the SyncModule.

        Args:
            parent (Any): Parent client or module.
            http_client (Any): HTTP client for making requests.
        """
        super().__init__(parent, http_client)


class AsyncModule(BaseModule):
    """Abstract base class for asynchronous modules in the Aisberg framework."""

    def __init__(self, parent: "AisbergAsyncClient", http_client: "AsyncHttpClient"):
        """
        Initialize the AsyncModule.

        Args:
            parent (Any): Parent client or module.
            http_client (Any): HTTP client for making requests.
        """
        super().__init__(parent, http_client)



class APIError(Exception):
    pass


class AuthError(APIError):
    pass


class ToolExecutionError(Exception):
    """Exception levée lors de l'exécution d'un tool"""

    pass


class ToolNotFoundError(Exception):
    """Exception levée lorsqu'un tool n'est pas trouvé"""

    def __init__(self, tool_name: str):
        super().__init__(f"Tool '{tool_name}' not found.")
        self.tool_name = tool_name


class UnspecifiedClassArgumentError(Exception):
    """Exception levée lorsqu'un argument requis n'est pas spécifié"""

    def __init__(self, argument_name: str):
        super().__init__(
            f"L'argument '{argument_name}' est requis mais n'a pas été spécifié."
        )
        self.argument_name = argument_name


class CollectionNotFoundError(Exception):
    """Exception levée lorsqu'une collection n'est pas trouvée"""

    def __init__(self, collection_name: str, group_id: str):
        super().__init__(
            f"La collection '{collection_name}' dans le groupe '{group_id}' n'a pas été trouvée."
        )
        self.collection_name = collection_name
        self.group_id = group_id



import os
import io
import sys
import mimetypes
import logging
from datetime import datetime
from abc import ABC, abstractmethod
from typing import Optional, Dict, Any, Union

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


# -------- Progress Bar --------
class ProgressPercentage:
    def __init__(self, file_size, object_name, bar_length=30, verbose=False):
        """
        Initialize the ProgressPercentage class.
        """
        self._file_size = file_size / 1024 / 1024  # MB
        self._seen_so_far = 0
        self._object_name = object_name
        self._bar_length = bar_length
        self._verbose = verbose

    def __call__(self, bytes_amount):
        if not self._verbose:
            return

        self._seen_so_far += bytes_amount / 1024 / 1024  # MB
        percentage = (self._seen_so_far / self._file_size) * 100
        bar_filled_length = int(self._bar_length * self._seen_so_far // self._file_size)
        bar = "█" * bar_filled_length + "-" * (self._bar_length - bar_filled_length)
        sys.stdout.write(
            f"\r{self._object_name}: |{bar}| {self._seen_so_far:.2f} MB / "
            f"{self._file_size:.2f} MB ({percentage:.2f}%)"
        )
        sys.stdout.flush()
        if self._seen_so_far >= self._file_size:
            sys.stdout.write("\n")
            sys.stdout.flush()


# -------- Base S3Module --------
class BaseS3Module(ABC):
    """
    Base abstraite pour modules S3.
    """

    def __init__(
        self,
        access_key_id: str,
        secret_access_key: str,
        endpoint_url: Optional[str] = None,
        region_name: str = "fr-par",
        verbose: bool = False,
    ):
        self._args = dict(
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            endpoint_url=endpoint_url,
            region_name=region_name,
        )
        self._verbose = verbose

    @abstractmethod
    def _get_client(self):
        """Returns the S3 client."""
        ...

    @staticmethod
    def _prepare_metadata(file, metadata=None):
        """Prepare metadata for the file.

        Args:
            file (str | io.BytesIO): path to the file or a BytesIO object.
            metadata (dict, optional): Additional metadata to enrich the file information.

        Returns:
            tuple: (file_size (int), metadata (dict)) with enriched file information.
        """
        if metadata is None:
            metadata = {}

        if isinstance(file, str):
            file_size = os.path.getsize(file)
            file_type, _ = mimetypes.guess_type(file)
            file_creation_time = datetime.fromtimestamp(
                os.path.getctime(file)
            ).isoformat()
            file_name = os.path.basename(file)
            metadata.update(
                {
                    "FileName": file_name,
                    "FileSize": f"{file_size / (1024 * 1024):.2f} MB",
                    "FileType": file_type or "application/octet-stream",
                    "CreationTime": file_creation_time,
                }
            )
        else:
            file.seek(0, io.SEEK_END)
            file_size = file.tell()
            file.seek(0)
            metadata.update(
                {
                    "FileSize": f"{file_size / (1024 * 1024):.2f} MB",
                    "FileType": "application/octet-stream",
                    "CreationTime": datetime.now().isoformat(),
                }
            )
        return file_size, metadata

    @staticmethod
    def _get_object_name(file, object_name):
        return object_name or (file if isinstance(file, str) else "default_object_name")

    @abstractmethod
    def upload_file(
        self,
        file: Union[str, io.BytesIO],
        bucket_name: str,
        object_name: Optional[str] = None,
        metadata: Optional[Dict[str, str]] = None,
    ) -> Any:
        """
        Uploads a file to the S3 bucket.

        Args:
            file (str or BytesIO): The file to upload. Can be a path to a file or a BytesIO object.
            bucket_name (str): The name of the S3 bucket.
            object_name (str, optional): The object name in S3. Defaults to filename if not provided.
            metadata (dict, optional): Additional metadata for the object.

        Returns:
            Any: True if the upload succeeded, else False or raises an exception.
        """
        ...

    @abstractmethod
    def download_file(
        self,
        bucket_name: str,
        object_name: str,
        save_to: Optional[str] = None,
    ) -> Any:
        """
        Downloads a file from the S3 bucket.

        Args:
            bucket_name (str): The name of the S3 bucket.
            object_name (str): The object name in S3.
            save_to (str, optional): Local path to save the file. If not provided, returns file content as BytesIO.

        Returns:
            Any: True if the file was saved locally, else a BytesIO object with file content, or False on failure.
        """
        ...

    @abstractmethod
    def delete_file(self, bucket_name: str, object_name: str) -> Any:
        """
        Deletes a file from the S3 bucket.

        Args:
            bucket_name (str): The name of the S3 bucket.
            object_name (str): The object name in S3.

        Returns:
            Any: True if the file was deleted, False otherwise.
        """
        ...

    @abstractmethod
    def list_files(self, bucket_name: str) -> Any:
        """
        Lists all files in the specified S3 bucket.

        Args:
            bucket_name (str): The name of the S3 bucket.

        Returns:
            Any: List of object names (keys) in the bucket, or an empty list on failure.
        """
        ...

    @abstractmethod
    def list_buckets(self) -> Any:
        """
        Lists all buckets accessible by the credentials.

        Returns:
            Any: List of bucket names, or an empty list on failure.
        """
        ...


# -------- Sync --------
class SyncS3Module(BaseS3Module):
    def _get_client(self):
        try:
            import boto3

            return boto3.client("s3", **self._args)
        except ImportError:
            logger.error(
                "boto3 is not installed. Please install it to use S3 functionalities."
            )
            raise ImportError("boto3 is required for S3 operations.")
        except Exception as e:
            logger.error(f"Failed to create S3 client: {e}")
            raise RuntimeError("Failed to create S3 client.") from e

    def upload_file(self, file, bucket_name, object_name=None, metadata=None) -> bool:
        s3 = self._get_client()
        object_name = self._get_object_name(file, object_name)
        file_size, metadata = self._prepare_metadata(file, metadata)
        progress_callback = ProgressPercentage(
            file_size, object_name, verbose=self._verbose
        )
        from boto3.s3.transfer import TransferConfig

        config = TransferConfig(use_threads=False)
        extra_args = {"Metadata": metadata}

        try:
            if isinstance(file, str):
                s3.upload_file(
                    file,
                    bucket_name,
                    object_name,
                    Config=config,
                    Callback=progress_callback,
                    ExtraArgs=extra_args,
                )
            else:
                s3.upload_fileobj(
                    file,
                    bucket_name,
                    object_name,
                    Config=config,
                    Callback=progress_callback,
                    ExtraArgs=extra_args,
                )
        except Exception as e:
            logger.error(f"Upload failed: {e}")
            return False
        return True

    def download_file(self, bucket_name, object_name, save_to=None):
        s3 = self._get_client()
        from boto3.s3.transfer import TransferConfig

        try:
            obj = s3.head_object(Bucket=bucket_name, Key=object_name)
            file_size = obj["ContentLength"]
            progress_callback = ProgressPercentage(
                file_size, object_name, verbose=self._verbose
            )
            config = TransferConfig(use_threads=False)

            if self._verbose:
                print(
                    f"Downloading {object_name} from bucket {bucket_name} "
                    f"({file_size / (1024 * 1024):.2f} MB)"
                )

            if save_to:
                s3.download_file(
                    bucket_name,
                    object_name,
                    save_to,
                    Config=config,
                    Callback=progress_callback,
                )
                return True
            else:
                file_obj = io.BytesIO()
                s3.download_fileobj(
                    bucket_name,
                    object_name,
                    file_obj,
                    Config=config,
                    Callback=progress_callback,
                )
                file_obj.seek(0)
                return file_obj
        except Exception as e:
            logger.error(f"Download failed: {e}")
            return False

    def delete_file(self, bucket_name, object_name):
        s3 = self._get_client()
        try:
            s3.delete_object(Bucket=bucket_name, Key=object_name)
            return True
        except Exception as e:
            logger.error(f"Delete failed: {e}")
            return False

    def list_files(self, bucket_name):
        s3 = self._get_client()
        try:
            response = s3.list_objects_v2(Bucket=bucket_name)
            return [obj["Key"] for obj in response.get("Contents", [])]
        except Exception as e:
            logger.error(f"List files failed: {e}")
            return []

    def list_buckets(self):
        s3 = self._get_client()
        try:
            response = s3.list_buckets()
            return [bucket["Name"] for bucket in response.get("Buckets", [])]
        except Exception as e:
            logger.error(f"List buckets failed: {e}")
            return []



from .config import settings
import httpx
from typing import Dict, Callable

from .modules import (
    AsyncChatModule,
    AsyncCollectionsModule,
    AsyncEmbeddingsModule,
    AsyncMeModule,
    AsyncModelsModule,
    AsyncWorkflowsModule,
    ToolsModule,
    AsyncDocumentsModule,
    SyncS3Module,
)


class AisbergAsyncClient:
    def __init__(self, base_url=None, api_key=None, timeout=None):
        self.api_key = api_key or settings.aisberg_api_key
        self.base_url = base_url or settings.aisberg_base_url
        self.timeout = timeout or settings.timeout
        self.tool_registry: Dict[str, Callable] = {}
        self._client = httpx.AsyncClient(
            base_url=self.base_url,
            timeout=self.timeout,
            headers={"Authorization": f"Bearer {self.api_key}"},
        )

        # Modules
        self.tools = ToolsModule(self)
        self.chat = AsyncChatModule(self, self._client)
        self.models = AsyncModelsModule(self, self._client)
        self.workflows = AsyncWorkflowsModule(self, self._client)
        self.me = AsyncMeModule(self, self._client)
        self.collections = AsyncCollectionsModule(self, self._client)
        self.embeddings = AsyncEmbeddingsModule(self, self._client)
        self.documents = AsyncDocumentsModule(self, self._client)
        self._s3 = SyncS3Module(
            settings.s3_access_key_id,
            settings.s3_secret_access_key,
            settings.s3_endpoint,
        )

    async def initialize(self):
        """
        Initialise le client asynchrone.
        Cette méthode est appelée pour s'assurer que le client est prêt à être utilisé.
        """
        await self._validate_api_key()
        return self

    async def _validate_api_key(self):
        """
        Valide la clé API en effectuant une requête à l'API.
        """
        try:
            await self.me.info()
        except httpx.ConnectTimeout as e:
            raise ConnectionError(
                f"Le host {self.base_url} n'est pas accessible. Vérifiez votre connexion réseau ou l'URL de l'API."
            ) from e
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 401:
                raise ValueError(
                    "Clé API invalide ou expirée. Veuillez vérifier votre clé API."
                ) from e
            elif e.response.status_code == 403:
                raise PermissionError(
                    "Accès interdit. Vérifiez vos permissions pour utiliser l'API."
                ) from e
            else:
                raise ValueError(
                    f"Erreur lors de la validation de la clé API: {e.response.text}"
                ) from e
        except Exception:
            raise ValueError(
                f"Clé API invalide/expirée ou le host {self.base_url} n'est pas accessible."
            )

    async def close(self):
        """
        Ferme le client.
        """
        await self._client.aclose()

    async def __aenter__(self):
        """
        Enter the context manager, returning the client instance.

        Example:
            async with AisbergAsyncClient() as client:
                # Use the client here
                pass

        Returns:
            AisbergAsyncClient: L'instance du client Aisberg.
        """
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        """
        Exit the context manager, closing the client.
        Example:
            async with AisbergAsyncClient() as client:
                # Use the client here
                pass
        """
        await self.close()

    def __repr__(self):
        return f"<AisbergAsyncClient base_url={self.base_url}>"

    def __str__(self):
        return f"AisbergAsyncClient(base_url={self.base_url})"



from typing import List, Union, Optional
from abc import ABC

import json


from ..models.collections import (
    GroupCollections,
    Collection,
    CollectionDetails,
    CollectionDataset,
)

from abc import abstractmethod
from ..abstract.modules import SyncModule, AsyncModule
from ..api import endpoints, async_endpoints
from ..models.requests import HttpxFileField
from io import BytesIO


class AbstractCollectionsModule(ABC):
    def __init__(self, parent, client):
        self._parent = parent
        self._client = client

    @abstractmethod
    def list(self) -> List[GroupCollections]:
        """
        Get a list of available collections. Collections are grouped by your belonging groups.

        Returns:
            List[GroupCollections]: A list of available collections.

        Raises:
            ValueError: If no collections are found.
            Exception: If there is an error fetching the collections.
        """
        ...

    @abstractmethod
    def get_by_group(self, group_id: str) -> List[Collection]:
        """
        Get collections by group ID.

        Args:
            group_id (str): The ID of the group for which to retrieve collections.

        Returns:
            List[Collection]: A list of collections for the specified group.

        Raises:
            ValueError: If no collections are found for the specified group ID.
            Exception: If there is an error fetching the collections.
        """
        ...

    @abstractmethod
    def details(self, collection_id: str, group_id: str) -> CollectionDetails:
        """
        Get details of a specific collection.

        Args:
            collection_id (str): The ID of the collection to retrieve.
            group_id (str): The ID of the group to which the collection belongs.

        Returns:
            CollectionDetails: The details of the specified collection.

        Raises:
            ValueError: If the specified collection is not found.
        """
        ...

    @abstractmethod
    def delete(self, name: str, **kwargs) -> bool:
        """
        Delete a collection by name and group ID.

        Args:
            name (str): The name of the collection to delete.
            **kwargs: Additional keyword arguments, such as group ID.

        Returns:
            bool: True if the deletion was successful, False otherwise.

        Raises:
            ValueError: If the collection could not be deleted.
            Exception: If there is an error during the deletion process.
        """
        ...

    @abstractmethod
    def create(
        self,
        name: str,
        data: Union[dict, CollectionDataset, str],
        embedding_model: Optional[str] = "BAAI/bge-m3",
        normalize: bool = False,
        **kwargs,
    ) -> CollectionDetails:
        """
        Create a new collection.

        Args:
            name (str): The name of the collection to create.
            data (Union[dict, CollectionDataset, str]): The data to insert into the collection.
                Can be a Dict, aCollectionDataset object or a string representing the file path.
            embedding_model (Optional[str]): The embedding model to use for the collection.
                Defaults to "BAAI/bge-m3".
            normalize (bool): Whether to normalize the data before inserting it into the collection. Defaults to False.
            **kwargs: Additional keyword arguments, such as group ID.

        Returns:
            CollectionDetails: The details of the created collection.
        Raises:
            ValueError: If the collection could not be created.
            Exception: If there is an error during the creation process.
        """
        ...

    @abstractmethod
    def insert_points(
        self,
        collection_name: str,
        data: Union[dict, CollectionDataset, str],
        normalize: bool = False,
        **kwargs,
    ) -> CollectionDetails:
        """
        Insert points into an existing collection. All existing points in the collection won't be deleted.
        This method is used to add new data to an existing collection without removing the previous data.

        Args:
            collection_name (str): The name of the collection to create.
            data (Union[dict, CollectionDataset, str]): The data to insert into the collection.
                Can be a Dict, aCollectionDataset object or a string representing the file path.
            normalize (bool): If collection already have points, the normalize parameter will be ignored. Defaults to False.
            **kwargs: Additional keyword arguments, such as group ID.

        Returns:
            CollectionDetails: The details of the created collection.
        Raises:
            ValueError: If the collection could not be created.
            Exception: If there is an error during the creation process.
        """
        ...

    @abstractmethod
    def delete_points(
        self,
        collection_name: str,
        points: List[str],
        **kwargs,
    ) -> CollectionDetails:
        """
        Delete points into an existing collection. Points with the specified IDs will be removed from the collection.

        Args:
            collection_name (str): The name of the collection to create.
            points (List[str]): The list of point IDs to delete from the collection.
            **kwargs: Additional keyword arguments, such as group ID.

        Returns:
            CollectionDetails: The details of the created collection.
        Raises:
            ValueError: If the collection could not be created.
            Exception: If there is an error during the creation process.
        """
        ...

    @abstractmethod
    def clear(
        self,
        collection_name: str,
        **kwargs,
    ) -> CollectionDetails:
        """
        Delete ALL points into an existing collection. All points will be removed from the collection. But the collection itself will not be deleted.
        So you will still be able to insert new points into the collection without creating a new one.

        Args:
            collection_name (str): The name of the collection to create.
            **kwargs: Additional keyword arguments, such as group ID.

        Returns:
            CollectionDetails: The details of the created collection.
        Raises:
            ValueError: If the collection could not be created.
            Exception: If there is an error during the creation process.
        """
        ...

    @staticmethod
    def _get_collections_by_group(
        collections: List[GroupCollections], group_id: str
    ) -> List[Collection]:
        for group in collections:
            if group.group == group_id:
                return group.collections
        raise ValueError("No collections found for group ID")

    @staticmethod
    def _data_to_httpx_file(
        data: Union[dict, CollectionDataset, str],
    ) -> HttpxFileField:
        """
        Prepare a JSON payload as a HTTPX file field (for multipart upload).

        Args:
            data (dict | CollectionDataset | str): The dataset as dict/obj or a path to a JSON file.

        Returns:
            HttpxFileField: List suitable for HTTPX multipart upload.
        """
        if isinstance(data, str):
            with open(data, "r", encoding="utf-8") as f:
                coll_dict = json.load(f)
            filename = data.split("/")[-1]
        elif isinstance(data, CollectionDataset):
            coll_dict = data if isinstance(data, dict) else data.model_dump()
            filename = "collection.json"
        elif isinstance(data, dict):
            if "chunks" in data and "metadata" in data:
                coll_dict = data
                filename = "collection.json"
            else:
                raise ValueError(
                    "data must be a dict with 'chunks' and 'metadata' keys"
                )
        else:
            raise ValueError(
                "data must be a dict, CollectionDataset, or file path string"
            )

        json_bytes = json.dumps(coll_dict, ensure_ascii=False).encode("utf-8")
        file_tuple = ("files", (filename, BytesIO(json_bytes), "application/json"))
        return [file_tuple]


class SyncCollectionsModule(SyncModule, AbstractCollectionsModule):
    def __init__(self, parent, client):
        SyncModule.__init__(self, parent, client)
        AbstractCollectionsModule.__init__(self, parent, client)

    def list(self) -> List[GroupCollections]:
        return endpoints.collections(self._client)

    def get_by_group(self, group_id: str) -> List[Collection]:
        collections = self.list()
        return self._get_collections_by_group(collections, group_id)

    def details(
        self, collection_id: str, group_id: Optional[str] = None
    ) -> CollectionDetails:
        points = endpoints.collection(self._client, collection_id, group_id)
        return CollectionDetails(
            name=collection_id,
            group=group_id,
            points=points if points else [],
        )

    def delete(self, name: str, **kwargs) -> bool:
        response = endpoints.delete_collection(self._client, name, **kwargs)
        if response is None:
            raise ValueError("Collection could not be deleted")
        return True

    def create(
        self,
        name: str,
        data: Union[dict, CollectionDataset, str],
        embedding_model: Optional[str] = "BAAI/bge-m3",
        normalize: bool = False,
        **kwargs,
    ) -> CollectionDetails:
        create = endpoints.create_collection(
            self._client, name, embedding_model, **kwargs
        )
        if create.message != "Creation started":
            raise ValueError("Collection could not be created")

        insert = endpoints.insert_points_in_collection(
            self._client,
            name,
            self._data_to_httpx_file(data),
            normalize,
            **kwargs,
        )
        if insert.message != f"Documents inserted in {name}.":
            raise ValueError("Points could not be inserted into the collection")

        return self.details(name, kwargs.get("group", None))

    def insert_points(
        self,
        collection_name: str,
        data: Union[dict, CollectionDataset, str],
        normalize: bool = False,
        **kwargs,
    ) -> CollectionDetails:
        insert = endpoints.insert_points_in_collection(
            self._client,
            collection_name,
            self._data_to_httpx_file(data),
            normalize,
        )
        if insert.message != f"Documents inserted in {collection_name}.":
            raise ValueError(
                f"Points could not be inserted into the collection : {insert.model_dump_json()}"
            )
        return self.details(collection_name, kwargs.get("group", None))

    def delete_points(
        self,
        collection_name: str,
        points: List[str],
        **kwargs,
    ) -> CollectionDetails:
        delete = endpoints.delete_points_in_collection(
            self._client,
            points,
            collection_name,
            **kwargs,
        )
        if (
            f'{len(points)} points deleted from collection "{collection_name}"'
            not in delete.message
        ):
            raise ValueError(
                f"Points could not be deleted from the collection : {delete.model_dump_json()}"
            )

        return self.details(collection_name, kwargs.get("group", None))

    def clear(
        self,
        collection_name: str,
        **kwargs,
    ) -> CollectionDetails:
        clear = endpoints.delete_all_points_in_collection(
            self._client,
            collection_name,
            **kwargs,
        )
        if (
            f'All points deleted from collection "{collection_name}" for group'
            not in clear.message
        ):
            raise ValueError(
                f"Points could not be deleted from the collection : {clear.model_dump_json()}"
            )
        return self.details(collection_name, kwargs.get("group", None))


class AsyncCollectionsModule(AsyncModule, AbstractCollectionsModule):
    def __init__(self, parent, client):
        AsyncModule.__init__(self, parent, client)
        AbstractCollectionsModule.__init__(self, parent, client)

    async def list(self) -> List[GroupCollections]:
        return await async_endpoints.collections(self._client)

    async def get_by_group(self, group_id: str) -> List[Collection]:
        collections = await self.list()
        return self._get_collections_by_group(collections, group_id)

    async def details(
        self, collection_id: str, group_id: Optional[str] = None
    ) -> CollectionDetails:
        points = await async_endpoints.collection(self._client, collection_id, group_id)

        return CollectionDetails(
            name=collection_id,
            group=group_id,
            points=points if points else [],
        )

    async def delete(self, name: str, **kwargs) -> bool:
        response = await async_endpoints.delete_collection(self._client, name, **kwargs)
        if response is None:
            raise ValueError("Collection could not be deleted")
        return True

    async def create(
        self,
        name: str,
        data: Union[dict, CollectionDataset, str],
        embedding_model: Optional[str] = "BAAI/bge-m3",
        normalize: bool = False,
        **kwargs,
    ) -> CollectionDetails:
        create = await async_endpoints.create_collection(
            self._client, name, embedding_model, **kwargs
        )
        if create.message != "Creation started":
            raise ValueError("Collection could not be created")

        insert = await async_endpoints.insert_points_in_collection(
            self._client,
            name,
            self._data_to_httpx_file(data),
            normalize,
            **kwargs,
        )
        if insert.message != f"Documents inserted in {name}.":
            raise ValueError("Points could not be inserted into the collection")

        return await self.details(name, kwargs.get("group", None))

    async def insert_points(
        self,
        collection_name: str,
        data: Union[dict, CollectionDataset, str],
        normalize: bool = False,
        **kwargs,
    ) -> CollectionDetails:
        insert = await async_endpoints.insert_points_in_collection(
            self._client,
            collection_name,
            self._data_to_httpx_file(data),
            normalize,
        )
        if insert.message != f"Documents inserted in {collection_name}.":
            raise ValueError(
                f"Points could not be inserted into the collection : {insert.model_dump_json()}"
            )

        return await self.details(collection_name, kwargs.get("group", None))

    async def delete_points(
        self,
        collection_name: str,
        points: List[str],
        **kwargs,
    ) -> CollectionDetails:
        delete = await async_endpoints.delete_points_in_collection(
            self._client,
            points,
            collection_name,
            **kwargs,
        )
        if (
            f'{len(points)} points deleted from collection "{collection_name}"'
            not in delete.message
        ):
            raise ValueError(
                f"Points could not be deleted from the collection : {delete.model_dump_json()}"
            )

        return await self.details(collection_name, kwargs.get("group", None))

    async def clear(
        self,
        collection_name: str,
        **kwargs,
    ) -> CollectionDetails:
        clear = await async_endpoints.delete_all_points_in_collection(
            self._client,
            collection_name,
            **kwargs,
        )
        if (
            f'All points deleted from collection "{collection_name}" for group'
            not in clear.message
        ):
            raise ValueError(
                f"Points could not be deleted from the collection : {clear.model_dump_json()}"
            )
        return await self.details(collection_name, kwargs.get("group", None))



from typing import Dict, List, Iterable
from pydantic import BaseModel, Field


class Property(BaseModel):
    type: str = Field(..., examples=["string"])
    description: str


class Parameters(BaseModel):
    """Schéma JSON Schema standard OpenAI pour les paramètres."""

    type: str = Field(default="object", frozen=True)
    properties: Dict[str, Property]
    required: List[str]


class Function(BaseModel):
    name: str
    description: str
    parameters: Parameters


class Tool(BaseModel):
    type: str = Field(default="function", frozen=True)
    function: Function


def make_tool(
    *,
    name: str,
    description: str,
    params: Dict[str, str],  # {"param": "description", ...}
) -> Tool:
    """Crée un objet `Tool` à partir des paramètres fournis.

    Args:
        name: Nom de la fonction.
        description: Description de la fonction.
        params: Dictionnaire des paramètres avec leur description.

    Returns:
        Un objet `Tool` prêt à être utilisé.
    """
    properties = {
        key: Property(type="string", description=desc)  # <-- adapt typage si besoin
        for key, desc in params.items()
    }
    return Tool(
        function=Function(
            name=name,
            description=description,
            parameters=Parameters(
                properties=properties,
                required=list(params.keys()),
            ),
        )
    )


def tools_to_payload(tools: Iterable[Tool]) -> List[Dict]:
    """
    Convertit des objets `Tool` (Pydantic) en payload brut..

    Args:
        tools: Un itérable d'instances `Tool`.
    Returns:
        Une liste de dicts sérialisables en JSON.
    """
    if not isinstance(tools, Iterable) or not tools:
        return []

    return [tool.model_dump(mode="json", exclude_none=True) for tool in tools]



from abc import ABC, abstractmethod
from ..api import endpoints, async_endpoints
from ..abstract.modules import SyncModule, AsyncModule
from ..models.documents import (
    FileObject,
    DocumentParserFileInput,
    ParsedDocument,
)
from typing import List
import json
from io import BytesIO
import logging

from ..models.requests import HttpxFileField

logger = logging.getLogger(__name__)


class AbstractDocumentsModule(ABC):
    def __init__(self, parent, client):
        self._parent = parent
        self._client = client

    @abstractmethod
    def parse(
        self, files: DocumentParserFileInput, **kwargs
    ) -> List[ParsedDocument]: ...

    def _get_parsed_files_from_s3(
        self, files: List[str], bucket_name: str
    ) -> List[ParsedDocument]:
        """
        Download and parse a list of files from an S3 bucket.

        Args:
            files (List[str]): List of file names to download from S3.
            bucket_name (str): Name of the S3 bucket.

        Returns:
            List[ParsedDocument]: Parsed documents as objects with content and metadata.

        Raises:
            Exception: If a file cannot be downloaded or parsed.
        """
        parsed_documents = []
        for file_name in files:
            if not file_name.endswith(".json"):
                if '"type": "error"' in file_name:
                    logger.error(f"[DOCUMENT PARSER] Parsing failed => {file_name}. ")
                continue

            logger.debug(f"Downloading file {file_name} from bucket {bucket_name}")
            # Download the file as a BytesIO
            doc_bytesio = self._parent._s3.download_file(bucket_name, file_name)
            try:
                buffer = doc_bytesio.getvalue()
                content_str = buffer.decode("utf-8")
                content_json = json.loads(content_str)
            finally:
                doc_bytesio.close()
            file_object = FileObject(name=file_name, buffer=buffer)
            parsed_documents.append(
                ParsedDocument(
                    content=content_json, metadata={"name": file_object.name}
                )
            )
        return parsed_documents

    @staticmethod
    def _prepare_files_payload(
        files: DocumentParserFileInput,
    ) -> HttpxFileField:
        """
        Prepares input files into a format compatible with HTTPX multipart uploads.

        Args:
            files (DocumentParserFileInput): Files to upload (see type for options).

        Returns:
            HttpxFileField: HTTPX-style list for multipart upload.

        Raises:
            TypeError: On unsupported type.
        """

        def to_file_tuple(item):
            # FileObject case
            if "FileObject" in globals() and isinstance(item, FileObject):
                content = item.buffer
                filename = item.name
            # (bytes, filename) tuple
            elif isinstance(item, tuple) and len(item) == 2:
                content, filename = item
            # bytes or BytesIO
            elif isinstance(item, (bytes, BytesIO)):
                content = item
                filename = "file"
            # str (filepath)
            elif isinstance(item, str):
                with open(item, "rb") as f:
                    content = f.read()
                filename = item.split("/")[-1]
            else:
                raise TypeError(
                    f"Unsupported file input type: {type(item)}. "
                    "Expected str, bytes, BytesIO, tuple, or FileObject."
                )
            # Normalize to BytesIO for HTTPX
            if isinstance(content, bytes):
                content = BytesIO(content)
            elif isinstance(content, BytesIO):
                content.seek(0)
            else:
                raise TypeError(
                    f"File content must be bytes or BytesIO, got {type(content)}"
                )
            return (filename, content)

        if isinstance(files, list):
            if len(files) == 0:
                raise ValueError("File list cannot be empty.")
            elif len(files) > 10:
                raise ValueError("Too many files provided. Maximum is 10.")

            normalized = [to_file_tuple(f) for f in files]
        else:
            normalized = [to_file_tuple(files)]

        # HTTPX format: [("files", (filename, fileobj, mimetype)), ...]
        httpx_files = [
            ("files", (filename, content, "application/octet-stream"))
            for filename, content in normalized
        ]
        return httpx_files


class SyncDocumentsModule(SyncModule, AbstractDocumentsModule):
    def __init__(self, parent, client):
        SyncModule.__init__(self, parent, client)
        AbstractDocumentsModule.__init__(self, parent, client)

    def parse(self, files, **kwargs) -> List[ParsedDocument]:
        output = endpoints.parse_documents(
            self._client,
            self._prepare_files_payload(files),
            **kwargs,
        )
        if output.message == "Files parsed successfully":
            return self._get_parsed_files_from_s3(output.parsedFiles, output.bucketName)
        else:
            raise ValueError(f"Error parsing files: {output.message}")


class AsyncDocumentsModule(AsyncModule, AbstractDocumentsModule):
    def __init__(self, parent, client):
        AsyncModule.__init__(self, parent, client)
        AbstractDocumentsModule.__init__(self, parent, client)

    async def parse(self, files, **kwargs) -> List[ParsedDocument]:
        output = await async_endpoints.parse_documents(
            self._client,
            self._prepare_files_payload(files),
            **kwargs,
        )
        if output.message == "Files parsed successfully":
            return self._get_parsed_files_from_s3(output.parsedFiles, output.bucketName)
        else:
            raise ValueError(f"Error parsing files: {output.message}")



import json

import httpx

from ..exceptions import CollectionNotFoundError
from ..models.chat import (
    LanguageModelInput,
    format_messages,
    ChatCompletionResponse,
    ChatCompletionChunk,
)
from typing import Optional, Generator, Union, List, Any

from ..models.collections import GroupCollections, PointDetails, ChunkingDictInput
from ..models.documents import DocumentParserResponse
from ..models.embeddings import (
    EncodingFormat,
    EncodingResponse,
    ChunksDataList,
    RerankerResponse,
)
from ..models.models import Model
from ..models.token import TokenInfo
from ..models.workflows import WorkflowDetails, Workflow
from ..utils import parse_chat_line, WorkflowLineParser
from ..requests.sync_requests import req, req_stream
from ..models.requests import AnyDict, AnyList, HttpxFileField


def models(client: httpx.Client) -> List[Model]:
    """
    Get the list of available models.
    """
    resp = req(client, "GET", "/v1/models", AnyDict).data
    if not resp or not isinstance(resp, list):
        raise ValueError("Invalid response format for models")
    return [Model.model_validate(item) for item in resp]


def workflows(client: httpx.Client) -> List[Workflow]:
    """
    Get the list of available workflows.
    """
    resp = req(client, "GET", "/workflow/light", AnyList)
    return [Workflow.model_validate(item) for item in resp.root]


def workflow(client: httpx.Client, workflow_id: str) -> WorkflowDetails:
    """
    Get details of a specific workflow.
    """
    try:
        resp = req(client, "GET", f"/workflow/details/{workflow_id}", WorkflowDetails)
        return resp
    except httpx.HTTPStatusError as e:
        if e.response.status_code == 404:
            raise ValueError(f"Workflow with ID {workflow_id} not found")
        raise e


def collections(client: httpx.Client) -> List[GroupCollections]:
    """
    Get the list of available collections.
    """
    resp = req(client, "GET", "/collections", AnyList)
    return [GroupCollections.model_validate(item) for item in resp.root]


def collection(
    client: httpx.Client, collection_id: str, group_id: Optional[str] = None
) -> List[PointDetails]:
    """
    Get details of a specific collection.
    """
    try:
        resp = req(client, "GET", f"/collections/{collection_id}/{group_id}", AnyList)
        return [PointDetails.model_validate(item) for item in resp.root]
    except httpx.HTTPStatusError as e:
        if e.response.status_code == 404:
            raise CollectionNotFoundError(
                collection_name=collection_id, group_id=group_id
            ) from e
        raise e


def create_collection(
    client: httpx.Client,
    name: str,
    model: str,
    group: Optional[str] = None,
):
    """
    Create a new collection with the specified name and optional group.
    """
    payload = {"collection_name": name, "embedding_model": model}
    if group is not None:
        payload["group"] = group

    return req(
        client,
        "POST",
        "/collections",
        AnyDict,
        json=payload,
    )


def delete_collection(
    client: httpx.Client,
    name: str,
    group: Optional[str] = None,
):
    """
    Delete a collection with the specified name and optional group.
    """
    payload = {"collections": [name]}
    if group is not None:
        payload["group"] = group

    return req(
        client,
        "DELETE",
        "/collections",
        AnyDict,
        json=payload,
    )


def insert_points_in_collection(
    client: httpx.Client,
    name: str,
    files: HttpxFileField,
    normalize: bool,
    chunking_dict: Optional[ChunkingDictInput] = None,
    group: Optional[str] = None,
):
    """
    Insert points into a collection with the specified name.
    """
    payload = {
        "chunking_dict": json.dumps({"method": "custom", "params": {}}),
        "normalize": normalize,
    }
    if group is not None:
        payload["group"] = group

    if chunking_dict is not None:
        payload["chunking_dict"] = chunking_dict.model_dump_json()

    return req(
        client,
        "POST",
        f"/collections/{name}",
        AnyDict,
        data=payload,
        files=files,
    )


def delete_points_in_collection(
    client: httpx.Client,
    points_ids: List[str],
    name: str,
    group: Optional[str] = None,
):
    """
    Delete points into a collection with the specified name.
    """
    payload = {
        "points": points_ids,
        "collection": name,
    }
    if group is not None:
        payload["group"] = group

    return req(
        client,
        "DELETE",
        "/collections/chunks",
        AnyDict,
        json=payload,
    )


def delete_all_points_in_collection(
    client: httpx.Client,
    name: str,
    group: Optional[str] = None,
):
    """
    Delete All points into a collection with the specified name.
    """
    payload = {
        "collection": name,
    }
    if group is not None:
        payload["group"] = group

    return req(
        client,
        "DELETE",
        "/collections/all/chunks",
        AnyDict,
        json=payload,
    )


def me(client: httpx.Client) -> TokenInfo:
    """
    Get the details of the current user.
    """
    return req(client, "GET", "/users/me", TokenInfo)


def chat(
    client: httpx.Client,
    input: LanguageModelInput,
    model: str = None,
    temperature: float = 0.7,
    tools: Optional[list] = None,
    group: Optional[str] = None,
    **kwargs,
) -> ChatCompletionResponse:
    """
    Send a chat message and get a response from an LLM endpoint.
    """
    if model is None:
        raise ValueError("Model must be specified")

    formatted_messages = format_messages(input)

    payload = {
        "model": model,
        "messages": formatted_messages,
        "temperature": temperature,
        "stream": False,
        **kwargs,
    }

    if group is not None:
        payload["group"] = group

    if tools is not None:
        payload["tools"] = tools

    return req(
        client,
        "POST",
        "/v1/chat/completions",
        ChatCompletionResponse,
        json=payload,
    )


def stream(
    client: httpx.Client,
    input: LanguageModelInput,
    model: str,
    temperature: float = 0.7,
    full_chunk: bool = True,
    group: Optional[str] = None,
    **kwargs,
) -> Generator[Union[str, ChatCompletionChunk, ChatCompletionResponse], None, None]:
    """
    Stream de complétions OpenAI.
    - Si `full_chunk` est True (défaut) : chaque yield est le JSON complet du chunk.
    - Sinon : on garde la compat ascendante → on ne yield que le delta.content + marquages.
    """
    formatted_messages = format_messages(input)

    payload = {
        "model": model,
        "messages": formatted_messages,
        "temperature": temperature,
        "stream": True,
        **kwargs,
    }

    if group is not None:
        payload["group"] = group

    for chunk in req_stream(
        client,
        "POST",
        "/v1/chat/completions",
        parse_line=lambda line: parse_chat_line(line, full_chunk=full_chunk),
        json=payload,
    ):
        data = ChatCompletionChunk.model_validate(chunk)

        if data is None:
            continue

        if full_chunk:
            yield data
        else:
            yield data.choices[0].delta.content if data.choices else ""


def embeddings(
    client: httpx.Client,
    input: str,
    model: str,
    encoding_format: EncodingFormat,
    normalize: bool,
    group: Optional[str] = None,
    **kwargs,
) -> EncodingResponse:
    """
    Get embeddings for a given input using the specified model.
    """
    payload = {
        "model": model,
        "input": input,
        "encoding_format": encoding_format,
        "normalize": normalize,
        **kwargs,
    }

    if group is not None:
        payload["group"] = group

    return req(
        client,
        "POST",
        "/v1/embeddings",
        EncodingResponse,
        json=payload,
    )


def retrieve(
    client: httpx.Client,
    query: str,
    collections_names: List[str],
    limit: int,
    score_threshold: float,
    filters: list,
    beta: float,
    group: Optional[str] = None,
    **kwargs,
) -> ChunksDataList:
    """
    Retrieve the most relevant documents based on the given query from specified collections.
    """
    data = {
        "query": query,
        "collections_names": collections_names,
        "limit": limit,
        "score": score_threshold,
        "filters": filters,
        "beta": beta,
        **kwargs,
    }

    if group is not None:
        data["group"] = group

    return req(
        client,
        "POST",
        "/collections/run/search",
        ChunksDataList,
        json=data,
    )


def rerank(
    client: httpx.Client,
    query: str,
    documents: List[str],
    model: str,
    top_n: int,
    return_documents: bool,
    group: Optional[str] = None,
    **kwargs,
) -> RerankerResponse:
    """
    Rerank a list of documents based on their relevance to a given query using the specified model.
    """
    payload = {
        "query": query,
        "documents": documents,
        "model": model,
        "top_n": top_n,
        "return_documents": return_documents,
        **kwargs,
    }

    if group is not None:
        payload["group"] = group

    return req(
        client,
        "POST",
        "/v1/rerank",
        RerankerResponse,
        json=payload,
    )


def run_workflow(
    client: httpx.Client,
    workflow_id: str,
    data: dict,
) -> Any:
    """
    Run a specific workflow with the provided data.
    """
    try:
        parser = WorkflowLineParser()
        for chunk in req_stream(
            client,
            "POST",
            f"/workflow/run/{workflow_id}",
            parse_line=parser,
            json=data,
        ):
            yield chunk
    except httpx.HTTPStatusError as e:
        if e.response.status_code == 404:
            raise ValueError(f"Workflow with ID {workflow_id} not found")
        raise e


def parse_documents(
    client: httpx.Client,
    files: HttpxFileField,
    group: Optional[str] = None,
    **kwargs,
) -> DocumentParserResponse:
    """
    Parse a single or multiple documents using the document parser endpoint.
    Returns the ID of the parsed document to be downloaded later from the S3 bucket.
    """
    payload = {**kwargs}
    if group is not None:
        payload["group"] = group

    response = req(
        client,
        "POST",
        "/document-parser/parsing/parse",
        DocumentParserResponse,
        files=files,
        data=payload,
    )
    return response



from .config import settings
import httpx
from typing import Dict, Callable
from .modules import (
    SyncChatModule,
    SyncCollectionsModule,
    SyncEmbeddingsModule,
    SyncMeModule,
    SyncModelsModule,
    SyncWorkflowsModule,
    ToolsModule,
    SyncDocumentsModule,
    SyncS3Module,
)


class AisbergClient:
    def __init__(self, base_url=None, api_key=None, timeout=None):
        self.api_key = api_key or settings.aisberg_api_key
        self.base_url = base_url or settings.aisberg_base_url
        self.timeout = timeout or settings.timeout

        if not self.base_url or not self.api_key:
            raise ValueError(
                "L'URL de base et la clé API doivent être définies. "
                "Utilisez les variables d'environnement AISBERG_API_KEY et AISBERG_BASE_URL ou passez-les lors de l'initialisation du client."
            )

        self.tool_registry: Dict[str, Callable] = {}
        self._client = httpx.Client(
            base_url=self.base_url,
            timeout=self.timeout,
            headers={"Authorization": f"Bearer {self.api_key}"},
        )

        # Modules
        self.tools = ToolsModule(self)
        self.chat = SyncChatModule(self, self._client)
        self.models = SyncModelsModule(self, self._client)
        self.workflows = SyncWorkflowsModule(self, self._client)
        self.me = SyncMeModule(self, self._client)
        self.collections = SyncCollectionsModule(self, self._client)
        self.embeddings = SyncEmbeddingsModule(self, self._client)
        self.documents = SyncDocumentsModule(self, self._client)
        self._s3 = SyncS3Module(
            settings.s3_access_key_id,
            settings.s3_secret_access_key,
            settings.s3_endpoint,
        )

        # Validate API key
        self._validate_api_key()

    def _validate_api_key(self):
        """
        Valide la clé API en effectuant une requête à l'API.
        """
        try:
            self.me.info()
        except httpx.ConnectTimeout as e:
            raise ConnectionError(
                f"Le host {self.base_url} n'est pas accessible. Vérifiez votre connexion réseau ou l'URL de l'API."
            ) from e
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 401:
                raise ValueError(
                    "Clé API invalide ou expirée. Veuillez vérifier votre clé API."
                ) from e
            elif e.response.status_code == 403:
                raise PermissionError(
                    "Accès interdit. Vérifiez vos permissions pour utiliser l'API."
                ) from e
            else:
                raise ValueError(
                    f"Erreur lors de la validation de la clé API: {e.response.text}"
                ) from e
        except Exception as e:
            raise ValueError(
                f"Clé API invalide/expirée ou le host {self.base_url} n'est pas accessible. Erreur: {str(e)}"
            )

    def close(self):
        """
        Ferme le client.
        """
        self._client.close()

    def __enter__(self):
        """
        Enter the context manager, returning the client instance.

        Example:
            with AisbergClient() as client:
                # Use the client here
                pass

        Returns:
            AisbergClient: L'instance du client Aisberg.
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Exit the context manager, closing the client.
        Example:
            with AisbergClient() as client:
                # Use the client here
                pass
        """
        self.close()

    def __repr__(self):
        return f"<AisbergClient base_url={self.base_url}>"

    def __str__(self):
        return f"AisbergClient(base_url={self.base_url})"



from typing import Union
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    # Variables attendues
    # -- API --
    aisberg_api_key: Union[str, None] = None
    aisberg_base_url: Union[str, None] = None
    timeout: int = 30

    # -- S3 --
    s3_access_key_id: Union[str, None] = None
    s3_secret_access_key: Union[str, None] = None
    s3_endpoint: Union[str, None] = None

    # Config Pydantic
    model_config = SettingsConfigDict(
        env_file=".env", env_file_encoding="utf-8", extra="ignore"
    )


settings = Settings()



from typing import List, Optional, Union
from pydantic import BaseModel


class Collection(BaseModel):
    name: str


class GroupCollections(BaseModel):
    group: str
    collections: List[Collection]


class CollectionDataset(BaseModel):
    chunks: List[str]
    metadata: Optional[dict] = []


class CollectionCreateResponse(BaseModel):
    message: Optional[str] = None


class ChunkingDictInput(BaseModel):
    method: Optional[str] = "custom"
    params: Optional[dict] = {}


# Modèle plus structuré pour payload
class Payload(BaseModel):
    method: Optional[str] = None
    norm: Optional[Union[str, bool]] = None
    filetype: Optional[str] = None
    filename: Optional[str] = None
    dense_encoder: Optional[str] = None
    Category: Optional[str] = None
    text: Optional[str] = None
    timestamp: Optional[str] = None
    collection_name: Optional[str] = None
    sparse_encoder: Optional[str] = None


class PointDetails(BaseModel):
    id: str
    payload: Payload


class CollectionDetails(BaseModel):
    name: str
    group: Optional[str] = None
    points: List[PointDetails]



from abc import ABC, abstractmethod
from typing import Optional, Union, Generator, AsyncGenerator, List, Awaitable

from ..models.chat import (
    LanguageModelInput,
    HumanMessage,
    ChatCompletionResponse,
    AIMessage,
    ToolMessage,
    ChatCompletionChunk,
)
from ..models.tools import Tool, tools_to_payload
from ..exceptions import ToolExecutionError
import json
from ..api import async_endpoints, endpoints
from ..abstract.modules import AsyncModule, SyncModule
import logging

logger = logging.getLogger(__name__)


class AbstractChatModule(ABC):
    """
    Abstract base class for chat modules.
    Handles common logic for processing tool calls in conversational LLM flows.
    """

    def __init__(self, parent, http_client):
        """
        Initialize the AbstractChatModule.

        Args:
            parent: Parent client instance.
            http_client: HTTP client for making requests.
        """
        self._parent = parent
        self._client = http_client

    @abstractmethod
    def complete(
        self,
        input: LanguageModelInput,
        model: str,
        temperature: float = 0.7,
        tools: Optional[List[Tool]] = None,
        auto_execute_tools: bool = False,
        **kwargs,
    ) -> Union[ChatCompletionResponse, "Awaitable[ChatCompletionResponse]"]:
        """
        Sends a chat message (or list of messages) to the model and returns a structured response.
        Supports tool usage and optional automatic tool execution.

        Args:
            input (LanguageModelInput): Message(s) to send to the model
            model (str): Identifier of the language model to use.
            temperature (float): Sampling temperature for response creativity (0.0–1.0, default: 0.7).
            tools (Optional[list]): List of tools that can be invoked by the model.
            auto_execute_tools (bool): If True, automatically executes tool calls generated by the model (default: False).
            **kwargs: Additional arguments passed to the API.

        Returns:
            ChatCompletionResponse: Structured model response.
        """
        ...

    @abstractmethod
    def stream(
        self,
        input: LanguageModelInput,
        model: str,
        temperature: float = 0.7,
        tools: Optional[List[Tool]] = None,
        auto_execute_tools: bool = True,
        full_chunk: bool = True,
        **kwargs,
    ) -> Union[
        Generator[Union[str, ChatCompletionChunk, ChatCompletionResponse], None, None],
        AsyncGenerator[Union[str, ChatCompletionChunk, ChatCompletionResponse], None],
    ]:
        """
        Streams a chat completion response from the model. If tools are provided,
        makes an initial non-streaming call to determine tool invocations,
        executes them if auto_execute_tools is enabled, and then streams the final response.

        Args:
            input (LanguageModelInput): Message(s) to send to the model
            model (str): Identifier of the language model to use.
            temperature (float): Sampling temperature for response creativity (0.0–1.0, default: 0.7).
            tools (Optional[list]): List of tools that can be invoked by the model.
            auto_execute_tools (bool): If True, automatically executes tool calls generated by the model (default: False).
            full_chunk (bool): If True, yields full response objects; if False, yields only the content (default: True).
            **kwargs: Additional keyword arguments for the API call.

        Yields:
            Union[str, ChatCompletionChunk, ChatCompletionResponse]: Response chunks or final responses as they become available.
        """
        ...

    @abstractmethod
    def _handle_tool_calls(
        self,
        response: ChatCompletionResponse,
        input,
        model,
        temperature,
        stream: bool = False,
        full_chunk: bool = True,
        **kwargs,
    ):
        """
        Handles execution of tool calls found in a model response. Automatically executes the requested tools and
        integrates their results into the conversation, then continues the chat either in streaming or blocking mode.

        Args:
            response (ChatCompletionResponse): The initial model response containing tool calls.
            input: Original input messages.
            model: Model identifier.
            temperature: Sampling temperature.
            stream: If True, returns a streaming generator.
            full_chunk: (Streaming only) If True, yields full chunks.
            **kwargs: Additional arguments for the next API call.

        Returns:
            ChatCompletionResponse or AsyncGenerator: If stream is True, returns a generator yielding tool results and final response.
            Otherwise, returns a complete ChatCompletionResponse with tool results integrated.
        """
        ...

    def _build_tool_messages(
        self,
        response: ChatCompletionResponse,
        input: Union[LanguageModelInput, List[HumanMessage]],
    ):
        """
        Utility: Build the message list including tool calls and their results.
        """
        tool_calls = response.choices[0].message.tool_calls
        messages: list = (
            input if isinstance(input, list) else [HumanMessage(content=input)]
        )

        messages.append(
            AIMessage(
                content=response.choices[0].message.content or "",
                tool_calls=tool_calls,
            )
        )

        for tool_call in tool_calls or []:
            tool_name = tool_call.function.name
            tool_arguments_str = tool_call.function.arguments
            try:
                tool_arguments = (
                    json.loads(tool_arguments_str) if tool_arguments_str else {}
                )
                tool_result = self._parent.tools.execute(tool_name, tool_arguments)
                messages.append(
                    ToolMessage(
                        content=json.dumps(tool_result)
                        if not isinstance(tool_result, str)
                        else tool_result,
                        tool_call_id=tool_call.id,
                    )
                )
            except json.JSONDecodeError:
                logger.error(
                    f"Invalid JSON arguments provided for tool {tool_name}."
                    f" Arguments: {tool_arguments_str}"
                )
                messages.append(
                    ToolMessage(
                        content=f"Error: Invalid JSON arguments provided for tool {tool_name}.",
                        tool_call_id=tool_call.id,
                    )
                )
            except ToolExecutionError as e:
                logger.error(f"Error executing tool {tool_name}: {str(e)}")
                messages.append(
                    ToolMessage(
                        content=f"Error: {str(e)}",
                        tool_call_id=tool_call.id,
                    )
                )
            except Exception as e:
                logger.error(f"Unexpected error executing tool {tool_name}: {str(e)}")
                messages.append(
                    ToolMessage(
                        content=f"Error: Unexpected error executing tool {tool_name}.",
                        tool_call_id=tool_call.id,
                    )
                )

        return messages


class SyncChatModule(AbstractChatModule, SyncModule):
    """
    `ChatModule` is a synchronous module that provides a high-level interface for interacting with
    language models in a conversational format. It supports standard completions, streaming responses,
    and automatic tool execution within conversations. The module abstracts all communication with the backend API,
    providing both blocking and generator-based usage.
    """

    def __init__(self, parent, http_client):
        super().__init__(parent, http_client)
        SyncModule.__init__(self, parent, http_client)

    def complete(
        self,
        input: LanguageModelInput,
        model: str,
        temperature: float = 0.7,
        tools: Optional[list[Tool]] = None,
        auto_execute_tools: bool = False,
        **kwargs,
    ) -> ChatCompletionResponse:
        response = endpoints.chat(
            self._client, input, model, temperature, tools_to_payload(tools), **kwargs
        )

        if (
            auto_execute_tools
            and response.choices
            and response.choices[0].message.tool_calls
        ):
            return self._handle_tool_calls(
                input=input,
                response=response,
                model=model,
                temperature=temperature,
                stream=False,
                **kwargs,
            )

        return response

    def stream(
        self,
        input: LanguageModelInput,
        model: str,
        temperature: float = 0.7,
        tools: Optional[list[Tool]] = None,
        auto_execute_tools: bool = True,
        full_chunk: bool = True,
        **kwargs,
    ) -> Generator[Union[str, ChatCompletionChunk, ChatCompletionResponse], None, None]:
        if tools:
            resp = endpoints.chat(
                self._client,
                input,
                model,
                temperature,
                tools_to_payload(tools),
                **kwargs,
            )
            response = ChatCompletionResponse.model_validate(resp)

            if not auto_execute_tools:
                yield response if full_chunk else response.choices[0].message.content
                return

            yield from self._handle_tool_calls(
                response,
                input,
                model,
                temperature,
                stream=True,
                full_chunk=full_chunk,
                **kwargs,
            )
        else:
            for chunk in endpoints.stream(
                self._client,
                input,
                model,
                temperature,
                full_chunk=full_chunk,
                **kwargs,
            ):
                yield ChatCompletionChunk.model_validate(chunk) if full_chunk else chunk

    def _handle_tool_calls(
        self,
        response: ChatCompletionResponse,
        input,
        model,
        temperature,
        stream: bool = False,
        full_chunk: bool = True,
        **kwargs,
    ):
        tool_calls = response.choices[0].message.tool_calls

        if tool_calls is None:
            return response

        messages = self._build_tool_messages(response, input)

        if stream:
            return self.stream(
                messages, model, temperature, full_chunk=full_chunk, **kwargs
            )
        else:
            return self.complete(messages, model, temperature, **kwargs)


class AsyncChatModule(AbstractChatModule, AsyncModule):
    """
    `ChatModule` is an asynchronous module that provides a high-level interface for interacting with
    language models in a conversational format. It supports standard completions, streaming responses,
    and automatic tool execution within conversations. The module abstracts all communication with the backend API,
    providing both blocking and generator-based usage.
    """

    def __init__(self, parent, http_client):
        super().__init__(parent, http_client)
        AsyncModule.__init__(self, parent, http_client)

    async def complete(
        self,
        input: LanguageModelInput,
        model: str,
        temperature: float = 0.7,
        tools: Optional[list[Tool]] = None,
        auto_execute_tools: bool = False,
        **kwargs,
    ) -> ChatCompletionResponse:
        resp = await async_endpoints.chat(
            self._client, input, model, temperature, tools_to_payload(tools), **kwargs
        )
        response = ChatCompletionResponse.model_validate(resp)

        if (
            auto_execute_tools
            and response.choices
            and response.choices[0].message.tool_calls
        ):
            return await self._handle_tool_calls(
                input=input,
                response=response,
                model=model,
                temperature=temperature,
                stream=False,
                **kwargs,
            )

        return response

    async def stream(
        self,
        input: LanguageModelInput,
        model: str,
        temperature: float = 0.7,
        tools: Optional[list[Tool]] = None,
        auto_execute_tools: bool = True,
        full_chunk: bool = True,
        **kwargs,
    ) -> AsyncGenerator[Union[str, ChatCompletionChunk, ChatCompletionResponse], None]:
        if tools:
            resp = await async_endpoints.chat(
                self._client,
                input,
                model,
                temperature,
                tools_to_payload(tools),
                **kwargs,
            )
            response = ChatCompletionResponse.model_validate(resp)

            if not auto_execute_tools:
                if full_chunk:
                    yield response
                else:
                    yield response.choices[0].message.content
                return

            async for val in await self._handle_tool_calls(
                response,
                input,
                model,
                temperature,
                stream=True,
                full_chunk=full_chunk,
                **kwargs,
            ):
                yield val
        else:
            async for chunk in async_endpoints.chat_stream(
                self._client,
                input,
                model,
                temperature,
                full_chunk=full_chunk,
                **kwargs,
            ):
                yield ChatCompletionChunk.model_validate(chunk) if full_chunk else chunk

    async def _handle_tool_calls(
        self,
        response: ChatCompletionResponse,
        input,
        model,
        temperature,
        stream: bool = False,
        full_chunk: bool = True,
        **kwargs,
    ):
        tool_calls = response.choices[0].message.tool_calls

        if tool_calls is None:
            return response

        messages = self._build_tool_messages(response, input)

        if stream:
            return self.stream(
                messages, model, temperature, full_chunk=full_chunk, **kwargs
            )
        else:
            return await self.complete(messages, model, temperature, **kwargs)
